''' Starter code for imitation learning of cartpole based on PID. If you want to log data, use
the log function rather than the print function, so what is printed is also saved to the log
file. Feel free to edit any part of the given code to fit your written code.'''
import time
from math import asin
from math import atan
from math import acos
import numpy as np
import torch
from pid_solution import Pid
from pole import CartPole

LOG_FILE = 'imitate.log' # Logs will be appended every time the code is run.
MODEL_FILE = 'imitate.model' # Once your program successfully trains a network, this file will be written
def log(*args, **kwargs):
    if LOG_FILE:
        with open(LOG_FILE, 'a+') as f:
            print(*args, file=f, **kwargs)
    print(*args, **kwargs)

class SpeedMemory:
    def __init__(self):
        self.last_error = None

    def get_derror(self, error, dt):
        if self.last_error is None:
            self.last_error = error
        derror = (error - self.last_error) / dt
        self.last_error = error
        return derror

def getNormalizedTensor(s1, s2, s3, s4):
    s1 = s1 / 255
    s2 = s2 / 50
    return torch.tensor([[s1, s2, s3, s4]])

def nearest_action(actions, desired_action):
    # Step 4: Find the action that the expert would take
    best = None

    return best

class Net(torch.nn.Module):
    def __init__(self, in_size, out_size, hidden_size):
        super().__init__()
        # Step 2.1: build neural network structure here

        # The below line are the nodes which normalizes the output for the probability distribution
        self.softmax = torch.nn.LogSoftmax(dim=1)

    def forward(self, x):
        # Step 2.2: define neural network forward pass here

        # after computing your neural network, normalize the output to 1.
        out = self.softmax(out)
        return out

    def save(self, file_path):
        torch.save(self.state_dict(), file_path)

    def load(self, file_path):
        self.load_state_dict(torch.load(file_path))
total_good_losses = 0
required_good_losses = 10
if __name__ == '__main__':

    # Create the agent of our imitation learning algorithm, in our case a NN that will take the current state as
    # input and return a next course of action.
    agent = Net(4, None, None) # Step 2.3: Replace the Nones with your network architecture dimensions.
    # The first number is the dimension of our state
    # The second input is the number of actions you define
    # The third number is the number of nodes in each hidden layer (we assume they are all the same size)

    # agent.load(MODEL_FILE) # uncomment for video when you have alread trained the system

    actions = [] # Step 1: replace this with your list of possible actions (i.e. possible motor velocity outputs)

    learning_rate = 0.0005

    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.Adam(agent.parameters(), lr=learning_rate)

    # Set up cartpole and PID expert
    cartpole = CartPole()
    render = False # Set to True to always include rendering
    if render:
        cartpole.rendering = True
    dt = 0.02

    pid = Pid(0.5, 0, 0.1)

    encoder = SpeedMemory()

    i = 0
    epochs = 20 # number of epochs
    for eps in range(epochs):
        done = False

        observation = cartpole.reset()

        counts = 0
        avgloss = 0
        while counts < 400: # episodes per epoch
            i+= 1
            tic = time.perf_counter() # timer to maintain loop frequency

            #update pid
            pid.update_error(observation)

            # Calculate error based on the encoder
            encoderError = observation[2] / 2000
            dEncoderError = encoder.get_derror(encoderError, dt)

            # Normalize the state for easier learning
            state = getNormalizedTensor(pid.error, pid.changeInError, encoderError, dEncoderError)

            outputNN = agent(state) # Output from NN
            # Find the best action to take from actions based on the output of the NN
            actionNumber = outputNN.argmax(1).item() # Get best action number
            actionNN = actions[actionNumber] # Get best action

            outputPID = pid.action(dt) # Output from PID
            # Step 3: Find the closest available action to the output of the PID
            actionPID = nearest_action(actions, outputPID) # Closest action option to PID output
            target = torch.tensor([actionPID]) # Target from expert

            loss = loss_fn(outputNN, target) # calculate loss between network action and expert action (PID action)

            avgloss += loss.item() # record loss for printing
            loss.backward() # perform backprop to collect gradients
            optimizer.step() # perform one step of update with optimizer
            optimizer.zero_grad() # reset accumulated gradients to 0

            observation, done, r = cartpole.step(actionNN) # perform a step of simulation

            pid.previousError = pid.error # Set previous error to current error

            if i % 400 == 0: # print status every 400 iterations
                avgloss /= 400
                if avgloss < 0.3: # if the net is close enough to the PID, render the environment
                    cartpole.rendering = True
                    total_good_losses += 1
                    if total_good_losses >= required_good_losses:
                        agent.save(MODEL_FILE)
                        log('Training complete. Model saved to {}'.format(MODEL_FILE))
                        cartpole.close()
                        exit()
                else:
                    cartpole.rendering = render or False
                log('Iteration: {}   Loss: {}'.format(i,avgloss))
                avgloss = 0

            if done:
                observation = cartpole.reset()
                counts += 1
                # Reset PID here
                pid = Pid(0.5, 0, 0.1)
