''' Solution code for PID-controlled cartpole simulator. Feel free to edit any part of the given code to fit your written code.'''
import time
from math import asin
from math import atan
from math import acos
from pole import CartPole
class Pid:
    def __init__(self, Kp, Ki, Kd):
        self.Kp = Kp
        self.Ki = Ki
        self.Kd = Kd
        self.error = 0
        self.previousError = 0
        self.changeInError = 0
        self.cummulativeError = 0

    def action(self, dt):
        motorCommand = self.Kp*self.error+self.Kd*self.changeInError+self.Ki*self.cummulativeError
        return motorCommand

    def update_error(self, observation):
        lightError = observation[0] - observation[1]  # Step 1.1: Calculate  error based on light sensors
        dLightError = lightError - self.previousError  # Step 1.2: Calculate change in error based on light sensors
        iLightError = self.cummulativeError + lightError

        self.error = lightError  # set the class attributes
        self.changeInError = dLightError  # set the class attributes
        self.cummulativeError = iLightError

def sign(x):
    if x >= 0:
        return 1
    return -1

if __name__ == '__main__':
    cartpole = CartPole() # Set up cartpole according to pole.py
    cartpole.rendering = True # True displays the video
    observation = cartpole.reset() # Remember: observation is [left reading, right reading, encoder value]

    dt = 0.02 # Time step

    pid = Pid(0.5,0,0.2) # Step 3: Tune the PID controller by changing these numbers.

    done = False #set up the loop
    while not done: # loop until done

        tic = time.perf_counter() # timer to maintain loop frequency

        pid.update_error(observation)

        action = pid.action(dt) # Get motor velocity from PID controller

        observation, done, _ = cartpole.step(action) # Send motor command to plant

        pid.previousError = pid.error # Set previous error to current error

    cartpole.close() # close the display
