''' Starter code for policy improvement of cartpole. If you want to log data, use
the log function rather than the print function, so what is printed is also saved to the log
file. Feel free to edit any part of the given code to fit your written code.

Make sure that you finish the imitation learning section first, to bootstrap and make your learning
here faster!'''
import time
import numpy as np
import torch
from math import asin
from math import atan
from math import acos
from pid_solution import Pid
from imitation_starter import Net
from pole import CartPole

LOG_FILE = 'improve.log' # Logs will be appended every time the code is run.
MODEL_FILE = 'improved.model'
IMITATE_MODEL_FILE = 'imitate.model'
def log(*args, **kwargs):
    if LOG_FILE:
        with open(LOG_FILE, 'a+') as f:
            print(*args, file=f, **kwargs)
    print(*args, **kwargs)

class SpeedMemory:
    def __init__(self):
        self.last_error = None

    def get_derror(self, error, dt):
        if self.last_error is None:
            self.last_error = error
        derror = (error - self.last_error) / dt
        self.last_error = error
        return derror

def getNormalizedTensor(s1, s2, s3, s4):
    s1 = s1 / 255
    s2 = s2 / 50
    return torch.tensor([[s1,s2,s3,s4]])

max_count = 0
learned_count = 10 # We consider the agent 'fully trained' when the env is solved 10 times in a row
def improve_policy(agent, episode):
    '''Performs a step of policy improvement based on the given discounted_reward.

    Note that `episode` is a list of tuples of the form
        (discounted_reward, state, action, reward)

    The output of your network should be a probability distribution, and `action` should be the
    index of the action used to create the state/action/reward tuple.'''
    global max_count
    avgreward = [Gt for Gt,_,_,_ in episode]
    avgreward = torch.mean(torch.tensor(avgreward)).item()
    avgloss = []
    if len(episode) == cartpole.max_steps:
        episode_trim = episode[:-50]
    else:
        episode_trim = episode
    for discounted_reward, state, action, reward in episode_trim:
        probs = agent(state)
        loss = probs[0,action] * (discounted_reward - avgreward)
        avgloss.append( loss.item())
        loss.backward()

    total_reward = sum([r for _,_,_,r in episode])
    max_reward = cartpole.max_steps

    if total_reward >= max_reward:
        max_count += 1
        print(max_count)
        if max_count > learned_count:
            log('Fully trained!')
            agent.save(MODEL_FILE)
            cartpole.close()
            exit()
    else:
        max_count = 0

    log('Loss: {} Total reward:{}'.format(round(sum(avgloss)/len(avgloss),3),
        total_reward))

if __name__ == '__main__':

    agent = Net(4, 15, 32)  # Step 1.1: initialize your network with the same parameters as for imitate

    agent.load(IMITATE_MODEL_FILE)
    optimizer = torch.optim.Adam(agent.parameters(), lr=5e-5)

    actions = [] # Step 1.2: copy over your actions from the imitation code

    cartpole = CartPole()
    cartpole.rendering = True # Uncomment this line for rendering
    cartpole.harder = True # You must train the policy improvement code with this set to True!
    dt = 0.02

    previousError = 0
    encoder = SpeedMemory()

    discountedReward = 0; # starts at 0
    discount_factor = 0.9 # Recommended discount factor, but feel free to change/play with this

    i = 0
    iterations = 1000
    for eps in range(iterations):
        done = False

        observation = cartpole.reset()

        counts = 0
        done = False
        episode = []
        while not done:
            tic = time.perf_counter()

            lightError = observation[0] - observation[1]
            dLightError = lightError - previousError

            # Calculate error based on the encoder
            encoderError = observation[2] / 2000
            dEncoderError = encoder.get_derror(encoderError, dt)

            state = getNormalizedTensor(lightError, dLightError, encoderError, dEncoderError)
            agentState = torch.tensor(state)

            outputNN = agent(state) # Output from NN
            actionNumber = outputNN.argmax(1).item() # Get best action number
            action = actions[actionNumber] # Get best action

            i+= 1
            observation, done, reward = cartpole.step(action)

            episode.append((agentState, actionNumber, reward)) # Append tuple of s, a, r to episode

            previousError = lightError

        # Setp 2: Calculate discounted reward and ammend episode so that there are tuples of (discounted reward, s, a, r)
        depisode = [(0,)]
        # calculate depisode
        improve_policy(agent, depisode[:-1])
        optimizer.step()
        optimizer.zero_grad()

        observation = cartpole.reset()
        counts += 1
    agent.save('partial-'+MODEL_FILE)
    log('Performance benchmark not met, saving model file to {} anyway'.format('partial-'+MODEL_FILE))

    cartpole.close()
